unit MakeOverlayForm;

{
Revision history:

2000 May 28  1.0    First version
2000 May 29  1.0.2  Trap non-existant data files correctly
                    Integrate with WXtrack for file names
                    Add font choice, but not colour choice!
2000 May 31  1.0.4  Add 1 & 2 degree options
                    Save user settings (except font)
2000 Jun 01  1.0.6  Add zoom facility
                    Remove erroneous edge lines and annotation
                    Correct Antartica, handle the single wrap-around line
2000 Jun 03  1.1.0  First released version
                    Add font save/restore
2001 Feb 11  1.2.0  Add choice of line thickness
2001 Aug 21  1.3.0  Correct layout for large fonts
2002 Apr 04  1.4.0  Add option for no gridlines
                    Add option for no yellow country boundaries

Copyright  David J Taylor, Edinburgh
Web site:  www.satsignal.net
E-mail:    davidtaylor@writeme.com
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, ComCtrls, RegistryVariables, djtCheckBox;

type
  TFormMain = class (TForm)
    ButtonMake: TButton;
    ButtonClose: TButton;
    StatusBar: TStatusBar;
    RadioGroupGrid: TRadioGroup;
    GroupBoxOptions: TGroupBox;
    ButtonFont: TButton;
    CheckBoxGridOnTop: TCheckBox;
    CheckBoxShowScale: TCheckBox;
    FontDialog: TFontDialog;
    ScrollBox: TScrollBox;
    ImageDisplay: TImage;
    ButtonZoom: TButton;
    ProgressBar: TProgressBar;
    LabelProgress: TLabel;
    ButtonAbout: TButton;
    EditGridlineWidth: TEdit;
    EditBoundaryWidth: TEdit;
    UpDownGridlineWidth: TUpDown;
    UpDownBoundaryWidth: TUpDown;
    LabelGridlineWidth: TLabel;
    LabelBoundaryLineWidth: TLabel;
    CheckBoxShowBoundaries: TdjtCheckBox;
    procedure ButtonAboutClick (Sender: TObject);
    procedure ButtonCloseClick (Sender: TObject);
    procedure ButtonFontClick (Sender: TObject);
    procedure ButtonMakeClick (Sender: TObject);
    procedure ButtonZoomClick (Sender: TObject);
    procedure FormCreate (Sender: TObject);
    procedure FormDestroy (Sender: TObject);
    procedure FormShow (Sender: TObject);
    procedure ScrollBoxResize (Sender: TObject);
    procedure UpDownGridlineWidthClick (Sender: TObject;  Button: TUDBtnType);
    procedure UpDownBoundaryWidthClick (Sender: TObject;  Button: TUDBtnType);
  private
    { Private declarations }
    bmp: TBitmap;
    boundary_width: TRegistryInteger;
    cartography_filename: string;
    font_settings: TRegistryFont;
    grid_index: TRegistryInteger;
    gridline_width: TRegistryInteger;
    grid_on_top: TRegistryBoolean;
    show_scale: TRegistryBoolean;
    master_overlay_filename: string;
    WXtrack_exe: string;
    zoomed_in: boolean;
    procedure do_zoom;
    procedure generate_cartographic_data (const filename: string);
    procedure SetStatusText (const Value: string);
    property StatusText: string write SetStatusText;
  public
    { Public declarations }
  end;

var
  FormMain: TFormMain;

implementation

{$R *.DFM}

uses
  ShellAPI, Cartography, HelpAboutBox;


function FreePhysicalMemoryMB: DWORD;
var
  MemoryStatus: TMemoryStatus;
begin
  MemoryStatus.dwLength := SizeOf (MemoryStatus);
  GlobalMemoryStatus (MemoryStatus);
  // Convert result to MB
  Result := MemoryStatus.dwAvailPhys shr 20;
end;


procedure TFormMain.FormCreate (Sender: TObject);
begin
  // Detrmine where WXtrack lives, will get the full file name
  WXtrack_exe := ExtractFilePath (GetRegisteredAppPath ('WXtrack'));
  cartography_filename := WXtrack_exe + boundary_data_filename;
  master_overlay_filename := WXtrack_exe + master_ovl_bmp_filename;
  ImageDisplay.Picture.Bitmap := TBitmap.Create;
  bmp := ImageDisplay.Picture.Bitmap;
  bmp.PixelFormat := pf8bit;
  // Try and force the bitmap to use the default palette
  bmp.Palette := GetStockObject (DEFAULT_PALETTE);

  // Suggested font defaults
  with bmp.Canvas.Font do
    begin
    Size := 11;
    Style := [fsBold];
    Name := 'Arial';
    end;
  font_settings := TRegistryFont.CreateFont ('Font', bmp.Canvas.Font);

  // Restore user settings
  grid_index := TRegistryInteger.CreateValue ('GridSpacing', 3);
  RadioGroupGrid.ItemIndex := grid_index.Value;
  show_scale := TRegistryBoolean.CreateValue ('Annotation', False);
  CheckBoxShowScale.Checked := show_scale.Value;
  grid_on_top := TRegistryBoolean.CreateValue ('GridOnTop', False);
  CheckBoxGridOnTop.Checked := grid_on_top.Value;
  gridline_width := TRegistryInteger.CreateValue ('GridlineWidth', 1);
  UpDownGridlineWidth.Position := gridline_width.Value;
  boundary_width := TRegistryInteger.CreateValue ('BoundaryWidth', 1);
  UpDownBoundaryWidth.Position := boundary_width.Value;
end;


procedure TFormMain.FormDestroy (Sender: TObject);
begin
  grid_index.Value := RadioGroupGrid.ItemIndex;
  grid_index.Free;
  show_scale.Value := CheckBoxShowScale.Checked;
  show_scale.Free;
  grid_on_top.Value := CheckBoxGridOnTop.Checked;
  grid_on_top.Free;
  font_settings.GetFrom (bmp.Canvas.Font);
  font_settings.Free;
  gridline_width.Free;
  boundary_width.Free;
end;


procedure TFormMain.FormShow (Sender: TObject);
begin
  with ImageDisplay do
    begin
//    ShowMessage (IntToStr (Width) + ' x ' + IntToStr (Height));
    end;
end;


procedure TFormMain.ButtonCloseClick (Sender: TObject);
begin
  Close;
end;


procedure TFormMain.generate_cartographic_data (const filename: string);

// In the cartography data, the polygons and polylines are in 16-bit
// minutes, ranging from -10800 (180 W) ... 0 (0) ... +10800 (180 E)
//              and from  +5400  (90 N) ... 0 (0) ...  -5400  (90 S)

  procedure CMapLonLat (lon, lat: double;  var ix, iy: integer);
  begin
    if lon > 180 then lon := lon - 360;
    ix := Round (((lon + 180) * cartog_width) / 360);
    iy := Round ((90 - lat) * cartog_height / 180);
  end;

  procedure SetStatusText (const st: string);
  begin
    StatusBar.SimpleText := st;
    StatusBar.Repaint;
    Application.ProcessMessages;
  end;

  function vis (const lon, lat: integer): boolean;
  begin
    Result := ((lat mod 10) = 0) and ((lon mod 10) = 0);
  end;

  procedure do_grid;
  var
    interval: integer;
    lon, lat: integer;
    x, y: integer;
    show_scale: boolean;
    st: string;
    dx: integer;
    dy: integer;
  begin
    case RadioGroupGrid.ItemIndex of
      0: interval := 1;
      1: interval := 2;
      2: interval := 5;
      3: interval := 10;
      4: interval := 15;
      5: interval := 20;
      6: interval := 30;
      7: Exit;
    else
      interval := 10;
    end;
    show_scale := CheckBoxShowScale.Checked;
    with bmp.Canvas do
      begin
      Pen.Color := clGrid;
      Font.Color := clGrid;
      Pen.Width := gridline_width.Value;
      for lon := -180 div interval to +180 div interval do
        if (Abs (interval * lon) <> 180) then
          begin
          CMapLonLat (interval * lon, -89, x, y);
          MoveTo (x, y);
          CMapLonLat (interval * lon, +89, x, y);
          LineTo (x, y);
          if show_scale then
            begin
            for lat := -90 div interval to 90 div interval do
              if (Abs (interval * lat) <> 90) and
                  vis (interval * lon, interval * lat) then
                begin
                CMapLonLat (interval * lon, interval * lat, x, y);
                st := IntToStr (Abs (interval * lon));
                if lon < 0
                  then st := st + 'W'
                  else if lon > 0
                    then st := st + 'E'
                    else st := st + '';
                dx := TextWidth (st);
                dy := TextHeight (st);
                TextOut (x - dx div 2, y + dy div 2, st);
                end;
            end;
          end;
      for lat := -90 div interval to 90 div interval do
        if (Abs (interval * lat) <> 90) then
          begin
          CMapLonLat (-180, interval * lat, x, y);
          MoveTo (x, y);
          CMapLonLat (+180, interval * lat, x, y);
          LineTo (x, y);
          if show_scale then
            begin
            for lon := -180 div interval to +180 div interval do
              begin
              if (Abs (interval * lon) <> 180) and
                  vis (interval * lon, interval * lat) then
                begin
                CMapLonLat (interval * lon, interval * lat, x, y);
                st := IntToStr (Abs (interval * lat));
                if lat < 0
                  then st := st + 'S'
                  else if lat > 0
                    then st := st + 'N'
                    else st := st + '';
                dx := TextWidth (st + '  ');
                dy := TextHeight (st);
                TextOut (x - dx, y - dy div 2, st);
                end;
              end;
            end;
          end;
      end;
  end;

var
  cartog_list: TCartogList;

  procedure do_boundaries;
  type
    PPointArray = ^TPointArray;
    TPointArray = array [0..0] of TPoint;
  var
    t0: DWORD;
    poly: integer;
    percent: integer;
    num_points: integer;
    old_x, old_y: integer;
    point: integer;
    new_x, new_y: integer;
    dx, dy: integer;
    mid_y: integer;
    PVertices: PPointArray;
    poly_type: integer;
    border_colour: TColor;
  begin
    GetMem (PVertices, cartog_list.MaxPoints * SizeOf (TPoint));
    with cartog_list, bmp.Canvas do
      begin
      Pen.Width := boundary_width.Value;
      border_colour := clLand;
      if CheckBoxShowBoundaries.Checked then border_colour := clBorder;
      t0 := 0;
      for poly := 0 to Count - 1 do
        begin
        percent := poly * 100 div Count;
        ProgressBar.Position := percent;
        if (GetTickCount - t0) > 350 then
          begin
          bmp.Modified := True;
          ImageDisplay.Repaint;
          t0 := GetTickCount;
          end;
        with cartog_list.Lines [poly] do
          begin
          num_points := 0;
          old_x := -1;
          old_y := -1;
          for point := 0 to count - 1 do
            begin
            with PPoints^ [point] do
              begin
              CMapLonLat (x / 60, y / 60, new_x, new_y);
              dx := Abs (old_x - new_x);
              dy := Abs (old_y - new_y);
              if (dx > (9 * cartog_width div 10)) and (y / 60 < -82) then
                begin
                mid_y := (old_y + new_y) div 2;
                PVertices^ [num_points].x := old_x;
                PVertices^ [num_points].y := old_y;
                Inc (num_points);
                PVertices^ [num_points].x := cartog_width - 1;
                PVertices^ [num_points].y := mid_y;
                Inc (num_points);
                PVertices^ [num_points].x := cartog_width - 1;
                PVertices^ [num_points].y := cartog_height - 1;
                Inc (num_points);
                PVertices^ [num_points].x := 0;
                PVertices^ [num_points].y := cartog_height - 1;
                Inc (num_points);
                PVertices^ [num_points].x := 0;
                PVertices^ [num_points].y := mid_y;
                Inc (num_points);
                end;
              if (dx > 0) or  (dy > 0) or (point = count - 1) then
                begin
                PVertices^ [num_points].x := new_x;
                PVertices^ [num_points].y := new_y;
                Inc (num_points);
                old_x := new_x;
                old_y := new_y;
                end;
              end;
            end;

          // Do countries (0), state boundaries (4), and lakes (8)
          poly_type := flags and $FE;
          Pen.Color := border_colour;
          case poly_type of
            0: Brush.Color := clLand;
            4: Brush.Color := clLand;
            8: Brush.Color := clWater;
          end;
          case poly_type of
            0, 4, 8:
              if (flags and 1) = 1
                then Polyline (Slice (PVertices^, num_points))
                else Polygon (Slice (PVertices^, num_points));
          end;

          end;
        end;
      end;
    FreeMem (PVertices);
  end;

const
  mem_required = 6;
var
  Attributes: DWORD;
  mem: DWORD;
  reply: Word;
  saved_cursor: TCursor;
begin
  ButtonZoom.Enabled := False;

  // See if we can read in the points required from Countries.dat
  cartog_list := TCartogList.Create;
  cartog_list.LoadFromFile (cartography_filename);

  if cartog_list.Count = 0 then
    begin
    cartog_list.Free;  // No countries.dat file so give up silently
    if Application.MessageBox
        (PChar ('MakeOverlay cannot find the file ' +
         cartography_filename + '!' + #13 + 'Would you like to download ' +
         'the zipped Country and State boundary data from my Web site?' +
         #13 + #13 + country_data_url + ' (size 459KB)'),
         'WXtrack - Country and State boundary data missing',
         MB_YESNO + MB_ICONQUESTION) = ID_YES
    then
      ShellExecute (Handle, 'open', country_data_url, nil, nil, SW_SHOWNORMAL);
    Close;
    Exit;
    end;

  mem := FreePhysicalMemoryMB;
  if mem < mem_required then
    begin
    reply := MessageDlg ('Less than ' + IntToStr (mem_required) +
      ' MB physical memory available (' + IntToStr (mem) +
      ' MB found).  Generating the master overlay bitmap on this ' +
      'PC may take many minutes, so instead you can download the ' +
      'zipped overlay bitmap from the Web site (413KB).  Click ' +
      ' Help for more information on downloading.' + #13 + #13 +
      'OK to continue to generate the master overlay, or cancel?',
      mtWarning, [mbOK, mbCancel {, mbHelp}], 0);
    if reply <> mrOK then
      begin
      cartog_list.Free;
      Exit;
      end;
    end;

  saved_cursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  StatusText := 'Creating overlay master image, please wait...';

  StatusText := 'Overlay master image, allocating memory...';
  with bmp do
    begin
    Height := 0;
    Canvas.Brush.Color := clWater;
    Height := cartog_height;
    Width := cartog_width;
    end;

  StatusText := 'Overlay master image, generating country bondaries';

  with bmp.Canvas do
    begin
    SetPolyFillMode (Handle, WINDING);
    Pen.Style := psSolid;
    if not CheckBoxGridOnTop.Checked then do_grid;
    do_boundaries;
    if CheckBoxGridOnTop.Checked then do_grid;
    end;

  cartog_list.Free;
  StatusText := 'Overlay master image, saving image, please wait...';
  try
    bmp.SaveToFile (filename);
    Attributes := GetFileAttributes (PChar (filename));
    Attributes := Attributes + FILE_ATTRIBUTE_READONLY;
    SetFileAttributes (PChar (filename), Attributes);
    StatusText := 'Overlay master image saved to ' + filename;
  except
    StatusText := 'Overlay master image could not be saved';
    Attributes := GetFileAttributes (PChar (filename));
    if (Attributes and FILE_ATTRIBUTE_READONLY) = FILE_ATTRIBUTE_READONLY
    then
      try
        if MessageDlg ('Do you wish to overwrite the existing read-only file?',
                    mtConfirmation, [mbYes, mbNo], 0) = mrYes
        then
          begin
          Attributes := Attributes - FILE_ATTRIBUTE_READONLY;
          SetFileAttributes (PChar (filename), Attributes);
          StatusText := 'Overlay master image, saving image, please wait...';
          bmp.SaveToFile (filename);
          Attributes := Attributes + FILE_ATTRIBUTE_READONLY;
          SetFileAttributes (PChar (filename), Attributes);
          StatusText := 'Overlay master image saved to ' + filename;
          end
        else
          StatusText := 'Overlay master image save cancelled by user';
      except
        StatusText := 'Overlay master image could not be saved';
        ShowMessage ('Error saving overlay master image to ' + filename +
                     '.  If WXtrack is open, you must close it before ' +
                     'trying again.');
      end
    else
      ShowMessage ('Error saving overlay master image to ' + filename +
                   '.  If WXtrack is open, you must close it before ' +
                   'trying again.');
  end;
  ProgressBar.Position := 0;
  Screen.Cursor := saved_cursor;
  ButtonZoom.Enabled := True;
end;


procedure TFormMain.ButtonMakeClick (Sender: TObject);
begin
  generate_cartographic_data (master_overlay_filename);
end;


procedure TFormMain.SetStatusText (const Value: string);
begin
  StatusBar.SimpleText := Value;
  Application.ProcessMessages;
end;


procedure TFormMain.ButtonFontClick (Sender: TObject);
begin
  with FontDialog do
    begin
    Font := bmp.Canvas.Font;
    if Execute then
      begin
      bmp.Canvas.Font := Font;
      end;
    end;
end;


procedure TFormMain.ButtonZoomClick (Sender: TObject);
begin
  with ImageDisplay do
    if ButtonZoom.Caption = 'Zoom in'
    then
      begin
      ButtonZoom.Caption := 'Zoom out';
      zoomed_in := True;
      do_zoom;
      end
    else
      begin
      ButtonZoom.Caption := 'Zoom in';
      zoomed_in := False;
      do_zoom;
      end;
end;


procedure TFormMain.do_zoom;
begin
  with ImageDisplay do
    if zoomed_in
    then
      begin
      ImageDisplay.Stretch := False;
      AutoSize := False;
      Width := Picture.Bitmap.Width;
      Height := Picture.Bitmap.Height;
      ScrollBox.HorzScrollBar.Visible := True;
      ScrollBox.VertScrollBar.Visible := True;
      end
    else
      begin
      ImageDisplay.Stretch := False;
      AutoSize := False;
      Width := ScrollBox.Width - 4;
      Height := ScrollBox.Height - 4;
      ScrollBox.HorzScrollBar.Visible := False;
      ScrollBox.VertScrollBar.Visible := False;
      Stretch := True;
      end;
end;


procedure TFormMain.ScrollBoxResize (Sender: TObject);
begin
  do_zoom;
end;


procedure TFormMain.ButtonAboutClick (Sender: TObject);
begin
  AboutBoxDJT1.ShowModal;
end;


procedure TFormMain.UpDownGridlineWidthClick (Sender: TObject;
  Button: TUDBtnType);
begin
  gridline_width.Value := UpDownGridlineWidth.Position;
end;


procedure TFormMain.UpDownBoundaryWidthClick (Sender: TObject;
  Button: TUDBtnType);
begin
  boundary_width.Value := UpDownBoundaryWidth.Position;
end;

end.


